;;; pic-asm-mode.el --- mode for editing Microchip PIC assember code

;; Copyright (C) 2004 by Robert James Kaes <rjk@wormbytes.ca>
;;   All rights reserved.

;; Author: Robert James Kaes <rjk@wormbytes.ca>
;; Keywords: tools, languages

;;; Commentary:

;; This mode extends the asm-mode.el included with Emacs to support
;; assembler source written for a Microchip PIC micro-controller.
;; Currently most of the extension is font colouring, and handling
;; identifiers correctly when inserting a colon.
;;
;; See asm-mode.el for more information as to what the base mode
;; does.

;;; Code:

(defvar pic-asm-grammar-identifier "\\(\\.?[a-z_?@#][a-z_0-9.?@#]*\\)")
(defvar pic-asm-grammar-end-of-line "\\(?:[ \t]\\|$\\)")

(defconst pic-asm-font-lock-keywords
  (eval-when-compile
    ;;
    ;; Define all the font-locking patterns
    ;;
    (list
     ;; preprocessor and preprocessor-like syntax
     '("^[ \t]*\\(#?\\(?:E\\(?:LSE\\|NDIF\\)\\|IF\\)\\)\\(?:[ \t]\\|$\\)"
       (1 font-lock-builtin-face))
     (cons (concat "^[ \t]*\\(\\(?:#?IFN?DEF\\)\\|\\(?:#\\(?:UN\\)?DEFINE\\)\\)[ \t]+"
                   pic-asm-grammar-identifier
                   pic-asm-grammar-end-of-line)
           '((1 font-lock-builtin-face) (2 font-lock-type-face)))
     '("^[ \t]*\\(#?INCLUDE\\)[ \t]+\\([<\"]?[^<>\";\r\n]*[>\"]?\\)\\(?:[ \t]\\|$\\)"
       (1 font-lock-builtin-face) (2 font-lock-string-face t))
     
     ;; identifiers as labels
     (cons (concat "^" pic-asm-grammar-identifier ":?")
           font-lock-function-name-face)

     ;; PIC assembler instructions
     (cons (concat
            "[ \t]" (regexp-opt
                   '("ADDWF" "ADDWFC" "ANDWF" "CLRF" "COMF" "CPFSEQ"
                     "CPFSGT" "CPFSLT" "DECF" "DECFSZ" "DCFSNZ" "INCF"
                     "INCFSZ" "INFSNZ" "IORWF" "MOVF" "MOVFF" "MOVWF"
                     "MULWF" "NEGF" "RLCF" "RLNCF" "RRCF" "RRNCF" "SETF"
                     "SUBFWB" "SUBWF" "SUBWFB" "SWAPF" "TSTFSZ" "XORWF"
                     "BCF" "BSF" "BTFSC" "BTFSS" "BTG" "BC" "BN" "BNC"
                     "BNN" "BNOV" "BNZ" "BOV" "BRA" "BZ" "CALL" "CLRWDT"
                     "DAW" "GOTO" "NOP" "POP" "PUSH" "RCALL" "RESET"
                     "RETFIE" "RETLW" "RETURN" "SLEEP" "ADDLW" "ANDLW"
                     "IORLW" "LFSR" "MOVLB" "MOVLW" "MULLW" "SUBLW"
                     "XORLW") t)
            pic-asm-grammar-end-of-line)
           font-lock-keyword-face)
     '("[ \t]\\(TBL\\(?:RD\\|WT\\)\\(?:\\*[+-]?\\|\\+\\*\\)\\)\\(?:[ \t]\\|$\\)" (1 font-lock-keyword-face))
     
     ;; MPASM/GPASM instructions
     (cons (concat
            "[ \t]" (regexp-opt
                   '("BANKSEL" "BANKISEL" "CBLOCK" "CODE" "CONSTANT"
                     "DA" "DB" "DE" "DT" "DW" "DATA" "IDATA"
                     "END" "ENDC" "ENDM" "ENDW" "EQU" "ERROR" "ERRORLEVEL"
                     "EXTERN" "EXITM" "EXPAND" "NOEXPAND" "FILL" "GLOBAL"
                     "LIST" "NOLIST" "LOCAL" "MACRO" "MESSG" "ORG"
                     "PAGE" "PAGESEL" "PROCESSOR" "RADIX" "RES" "SET" "SPACE"
                     "SUBTITLE" "TITLE"
                     "UDATA" "UDATA_ACS" "UDATA_OVR" "UDATA_SHR"
                     "VARIABLE" "WHILE" "HIGH" "UPPER" "LOW") t)
            pic-asm-grammar-end-of-line)
           font-lock-builtin-face)
     (cons (concat
            "[ \t]" (regexp-opt
                      '("__BADRAM" "__CONFIG" "__IDLOCS" "__MAXRAM") t)
            "[ \t]")
           font-lock-builtin-face)
     )) "Default expressions to highlight in PIC-ASM mode.")

(require 'asm-mode)

(define-derived-mode pic-asm-mode asm-mode "PIC-Assembler"
  "Major mode for editing typical Microchip PIC assembler code.
Extends the standard `asm-mode' so see it's documation for more
information about what this mode supports.

The major differences are improved font colouring and a fix to
handle MPASM/GPASM indentifiers correctly.
"
  (setq font-lock-defaults '(pic-asm-font-lock-keywords nil t))
  (define-key asm-mode-map ":" 'pic-asm-colon))

;;
;; Copied this code from asm-mode.el since we need to override the regular
;; expression used to match a label
;;
(defun pic-asm-colon ()
  "Insert a colon; if it follows a label, delete the label's indentation."
  (interactive)
  (save-excursion
    (beginning-of-line)
    (if (looking-at (concat "[ \t]*" pic-asm-grammar-identifier))
	(delete-horizontal-space)))
  (insert ":")
  (newline)
  (tab-to-tab-stop)
  )

(provide 'pic-asm-mode)
